#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <iconv.h>
#include "libgsmat.h"
#include "gsm_internal.h"

unsigned long gsm_hex2int(char *hex, int len)
{
	int i,j;
	unsigned long res = 0;

	for (i=0; i<len; i++) {
		j = 0;
		if ((hex[i] >= 48) && (hex[i] <=57)) { /* '0' - '9' */
			j = hex[i] - 48;
		} else if ((hex[i] >= 65) && (hex[i] <=70)) { /* 'A' - 'F' */
			j = hex[i] - 55;
		} else if ((hex[i] >= 97) && (hex[i] <=102)) { /* 'a' - 'f' */
			j = hex[i] - 87;
		}
		/* j = [0-9,A-F,a-f] */
		res += j * (unsigned long)pow(16, len - i - 1);
	}
	return res;
}

static void gsm_dso2string(char *res, char *dso)
{
	int i;
	int len = strlen(dso);

	for (i=0; i < len; i+=2) {
		res[i] = dso[i+1];
		if ((dso[i] != 'F') && (dso[i] != 'f')) {
				res[i+1] = dso[i];
		} else {
				res[i+1] = '\0';
				return;
		} 
	}

	if (i >= len) {
		res[i] = '\0';
	}
}

static void gsm_convertNumber(char *res, char *dso, int toa)
{
	if (toa == 0x91) {
		res[0] = '+';
		gsm_dso2string(res+1, dso);
	} else {
		gsm_dso2string(res, dso);
	}
}

static void gsm_string2byte(unsigned char *res, char *in, int len)
{
	int i;

	for (i = 0; i < len ; i++) {
		res[i] = (unsigned char)gsm_hex2int(&in[i*2], 2);
	}
	res[len * 2] = '\0';
}

static char *gsm_get_dot_string(char *in, char *out, char flag)
{
	char *ret;
	int len;

	if (NULL == in && NULL == out) {
		return 0;
	}

	ret = strchr(in, flag);
	len = ret - in;
	if (ret != NULL) {
		if (len >= 255) {
			return NULL;
		}
		strncpy(out, in, len);
		out[len] = '\0';
		return ret + 1;
	}

	return NULL;
}

static void gsm_to8Bit(unsigned char in[], unsigned char out[], int len)
{
	int i = 0;
	int inputOffset = 0;
	int outputOffset = 0;

	while (inputOffset < len ) {
	out[outputOffset] = (in[inputOffset] & (unsigned char)(pow(2, 7-i)-1));
	if (i == 8) {
		out[outputOffset] = in[inputOffset-1] & 127;
		i = 1;
	} else {
		out[outputOffset] = out[outputOffset] << i;
		out[outputOffset] |= in[inputOffset-1] >> (8-i);
		inputOffset++;
		i++;
	}
		outputOffset++;
	}

	out[len] = '\0';
}

static inline char *pdu_get_smsc_len(char *body, char *smsc_len)
{
	char *pBegin, *pEnd;

	pBegin = body;
	pEnd = pBegin + 2;

	*smsc_len = (char)gsm_hex2int(pBegin, 2);

	return pEnd;
}

static inline char *pdu_get_smsc_toa(char *body, char *smsc_toa)
{
	char *pBegin, *pEnd;

	pBegin = body;
	pEnd = pBegin + 2;
		
	*smsc_toa = (char)gsm_hex2int(pBegin, 2);

	return pEnd;
}

static inline char *pdu_get_smsc_number(char *body, int smsc_len, char *smsc_number_dso)
{
	char *pBegin, *pEnd;
	
	int smsc_number_len = (smsc_len - 1) * 2;		
	pBegin = body;
	pEnd = pBegin + smsc_number_len;
	
	strncpy(smsc_number_dso, pBegin, smsc_number_len);
	smsc_number_dso[smsc_number_len] = '\0';

	return pEnd;
}

static inline char *pdu_get_first_octet(char *body, char *first_octet)
{
	char *pBegin, *pEnd;
	
	pBegin = body;
	pEnd = pBegin + 2;
		
	*first_octet = (char)gsm_hex2int(pBegin, 2);

	return pEnd;	
}

static inline char *pdu_get_tpoa_length(char *body, char *oa_length)
{
	char *pBegin, *pEnd;
	
	pBegin = body;
	pEnd = pBegin + 2;
		
	*oa_length = (char)gsm_hex2int(pBegin, 2);

	return pEnd;	
}

static inline char *pdu_get_tpoa_type(char *body, char *oa_type)
{
	char *pBegin, *pEnd;
	
	pBegin = body;
	pEnd = pBegin + 2;
		
	*oa_type = (char)gsm_hex2int(pBegin, 2);

	return pEnd;	
}

static inline char *pdu_get_tpoa_number(char *body, int oa_len, char *oa_number)
{
	char *pBegin, *pEnd;

	pBegin = body;

	if (oa_len % 2) {
		/* odd */
		oa_len++;
	}
	pEnd = pBegin + oa_len;

	strncpy(oa_number, pBegin, oa_len);
	oa_number[oa_len] = '\0';

	return pEnd;
}

static inline char *pdu_get_tppid(char *body, char *tppid)
{
	char *pBegin, *pEnd;
	
	pBegin = body;
	pEnd = pBegin + 2;
		
	*tppid = (char)gsm_hex2int(pBegin, 2);

	return pEnd;	
}

static inline char *pdu_get_tpdcs(char *body, char *tpdcs)
{
	char *pBegin, *pEnd;
	
	pBegin = body;
	pEnd = pBegin + 2;
		
	*tpdcs = (char)gsm_hex2int(pBegin, 2);

	return pEnd;	
}

static inline char *pdu_get_tpscts(char *body, char *tpscts)
{
	char *pBegin, *pEnd;
	int len = 14;
	
	pBegin = body;
	pEnd = pBegin + len;
	
	strncpy(tpscts, pBegin, len);
	tpscts[len] = '\0';

	return pEnd;
}

static inline char *pdu_get_tpudl(char *body, unsigned char *tpudl)
{
	char *pBegin, *pEnd;
	
	pBegin = body;
	pEnd = pBegin + 2;
		
	*tpudl = (unsigned char)gsm_hex2int(pBegin, 2);

	return pEnd;	
}

int gsm_decode_pdu(struct gsm_modul *gsm, char *pdu, struct gsm_sms_pdu_info *pdu_info)
{
	int len;

	char smsc_number_dso[32];
	char sender_number_dso[32];
	char timestamp_dso[16];
	char *p;
	
	len = strlen(pdu);
	if (len < 14) {
		gsm_error(gsm, "PDU too short\n");
		return -1;
	}

	p = pdu;
	
	memset(pdu_info, 0, sizeof(struct gsm_sms_pdu_info));

	/* Length of the SMSC information */
	p = pdu_get_smsc_len(p, (char *)&pdu_info->smsc_addr_len);

	if (pdu_info->smsc_addr_len > 0) {
		/* Type-of-address of the SMSC */
		p = pdu_get_smsc_toa(p, (char *)&pdu_info->smsc_addr_type);

		/* SMSC Number (in decimal semi-octets) */
		if ((pdu_info->smsc_addr_len - 1) * 2 >= sizeof(smsc_number_dso)) {
			gsm_error(gsm, "smsc number is too long\n");
			return -1;
		}
		p = pdu_get_smsc_number(p, pdu_info->smsc_addr_len, smsc_number_dso);
		gsm_convertNumber((char *)pdu_info->smsc_addr_number, smsc_number_dso, pdu_info->smsc_addr_type);
	} 

	/* PDU Type */
	p = pdu_get_first_octet(p, (char *)&pdu_info->first_octet);

	/* TP-OA */
	
	/* Address-Length, Length of the sender number */
	p = pdu_get_tpoa_length(p, (char *)&pdu_info->tp_oa.len);
	if (pdu_info->tp_oa.len > 0) {
		/* Type of address */
		p = pdu_get_tpoa_type(p, (char *)&pdu_info->tp_oa.type);

		/* Sender number (decimal semi-octets), with a trailing F */
		if (pdu_info->tp_oa.len + 1 >= sizeof(sender_number_dso)) {
			gsm_error(gsm, "Sender number is too long\n");
			return -1;
		}
		p = pdu_get_tpoa_number(p, pdu_info->tp_oa.len, sender_number_dso);
		gsm_convertNumber(pdu_info->tp_oa.number, sender_number_dso, pdu_info->tp_oa.type);
	}

	/* TP-PID: Protocol identifier */
	p = pdu_get_tppid(p, (char *)&pdu_info->tp_pid);    


	/* TP-DCS: Data coding scheme */
	p = pdu_get_tpdcs(p, (char *)&pdu_info->tp_dcs);    


	/* TP-SCTS: Time stamp (semi-octets */
	p = pdu_get_tpscts(p, timestamp_dso);
	gsm_dso2string(pdu_info->tp_scts, timestamp_dso);

	/* 
	 * TP-UDL: User data length, length of message. 
	 * The TP-DCS field indicated 7-bit data, so the length here is the number of septets (10). 
	 * If the TP-DCS field were set to indicate 8-bit data or Unicode, the length would be the number of octets (9).
	 */
	p = pdu_get_tpudl(p, &pdu_info->tp_udl);

	/* TP-UD */
	gsm_copy_string(pdu_info->tp_ud, p, sizeof(pdu_info->tp_ud));

	return 0;
}

int gsm_pdu2sm_event(struct gsm_modul *gsm, char *pdu) 
{
	struct gsm_sms_pdu_info pdu_info;
	int ret = 0;
	unsigned char user_data_7bit[320];
	unsigned char user_data_8bit[320];
	unsigned char udhi = 0;
	char buf[512];
	unsigned long header_len = 0;

	gsm_copy_string(gsm->ev.sm_received.pdu, pdu, sizeof(gsm->ev.sm_received.pdu));
	gsm_copy_string(buf, pdu, sizeof(buf));

//	gsm->debug=1;
	
	if (gsm->debug) {
		gsm_message(gsm, "SMS PDU = %s\n", pdu);
	}
	
	ret = gsm_decode_pdu(gsm, buf, &pdu_info);
	if (ret < 0) {
		return -1;
	}

	if (pdu_info.first_octet & GSM_SMS_PDU_FO_TP_UDHI) {
		udhi = 1;
		gsm_message(gsm, "User Data Header Indicator = %d\n", udhi);
	}

	if (gsm->debug) {
		gsm_message(gsm, "SMSC Length = %d\n", pdu_info.smsc_addr_len);
	}
	
	if (pdu_info.smsc_addr_len > 0) {
		if (gsm->debug) {
			gsm_message(gsm, "SMSC Type-of-address = %#x\n", pdu_info.smsc_addr_type);
			gsm_message(gsm, "SMSC Number = %s\n", pdu_info.smsc_addr_number);
		}
		gsm_copy_string(gsm->ev.sm_received.smsc, pdu_info.smsc_addr_number, sizeof(gsm->ev.sm_received.smsc));
	} else {
		gsm->ev.sm_received.smsc[0] = '\0';
	}

	if (gsm->debug) {
		gsm_message(gsm, "PDU First Octet = %#x\n", pdu_info.first_octet);
		gsm_message(gsm, "Sender number length = %d\n", pdu_info.tp_oa.len);
	}

	if (pdu_info.smsc_addr_len > 0) {
		if (gsm->debug) {
			gsm_message(gsm, "Sender Type-of-address = %#x\n", pdu_info.tp_oa.type);
			gsm_message(gsm, "Sender Number = %s\n", pdu_info.tp_oa.number);
		}
		gsm_copy_string(gsm->ev.sm_received.sender, pdu_info.tp_oa.number, sizeof(gsm->ev.sm_received.sender));
	} else {
		gsm->ev.sm_received.sender[0] = '\0';
	}

	if (gsm->debug) {
		gsm_message(gsm, "TP-PID = %#x\n", pdu_info.tp_pid);
		gsm_message(gsm, "TP-DCS = %#x\n", pdu_info.tp_dcs);
		gsm_message(gsm, "Timestamp = %s\n", pdu_info.tp_scts);
		gsm_message(gsm, "User data length = %d\n", pdu_info.tp_udl);
		gsm_message(gsm, "User data = %s\n", pdu_info.tp_ud);
	}
	gsm->ev.sm_received.len = pdu_info.tp_udl;

	/* TP-UD */
	if (udhi) {
		header_len = gsm_hex2int((char *)pdu_info.tp_ud, 2) + 1;
		gsm->ev.sm_received.len -= header_len;
	}

	gsm_string2byte(user_data_7bit, pdu_info.tp_ud + header_len * 2, pdu_info.tp_udl - header_len);
	if (pdu_info.tp_dcs == 0) {
		gsm_to8Bit(user_data_7bit, user_data_8bit, pdu_info.tp_udl - header_len);		
	} else {
		user_data_8bit[0] = '\0';
	}

	if (gsm->debug) {
		gsm_message(gsm, "User data 8bit = %s\n", user_data_8bit);
	}
	
	memset(gsm->ev.sm_received.text, 0, sizeof(gsm->ev.sm_received.text));
	if (pdu_info.tp_udl) {
		if (pdu_info.tp_dcs == 0x08) {
			sms_get_str((char *)user_data_7bit, gsm->ev.sm_received.len, gsm->ev.sm_received.text, 255);
		} else {
			strncpy(gsm->ev.sm_received.text, (char *)user_data_8bit, pdu_info.tp_udl - header_len);
		}
	}

	return 0;
}


static int code_convert(char *from_charset,char *to_charset,char *inbuf,size_t inlen,char *outbuf,size_t outlen)
{
	iconv_t cd;
	char **pin = &inbuf;
	char **pout = &outbuf;

	cd = iconv_open(to_charset,from_charset);
	if (cd==0) {
		return -1;
	}

	memset(outbuf,0,outlen);

	if (iconv(cd,pin,&inlen,pout,&outlen)==-1) {
		return -1;
	}

	iconv_close(cd);

	return outlen;
}

static int unicode_2_gb18030(char *inbuf,size_t inlen,char *outbuf,size_t outlen)
{
	return code_convert("unicode","gb18030",inbuf,inlen,outbuf,outlen);
}
static int gb18030_2_unicode(char *inbuf,size_t inlen,char *outbuf,size_t outlen)
{
	return code_convert("gb18030","unicode",inbuf,inlen,outbuf,outlen);
}
static int sms_decode_gb18030(char *in, size_t inlen, char *out)
{
	char tmp[512];
	char swap_char;
	int outleft;
	int i;
	int j;
	int k;
	int out_len;

	outleft = gb18030_2_unicode(in,inlen,tmp,512);
	if (outleft <= 0) {
		return -1;
	}

	out_len = 512 - outleft;
	for (i = 2; i < out_len; i=i+2) {
		swap_char = tmp[i];
		tmp[i - 2] = tmp[i + 1];
		tmp[i - 1] = swap_char;
	}

	out_len -= 2;

	for (k = 0, i = 0, j = 0; i < out_len; i ++) {
		if (0x1A == tmp[i]) {
			out[k++] = 0xFF;
			out[k++] = 0xFF;
			out[k++] = 0xFA;
			j += 2;
		} else if (0x1B == tmp[i]) {
			out[k++] = 0xFF;
			out[k++] = 0xFF;
			out[k++] = 0xFB;
			j += 2;
		} else {
			out[k++] = tmp[i];
		}
	}

	out_len += j;

	for (i = 0; i< out_len ; i ++) {
		out[i] = (char)(out[i]);
	}

	return out_len;
}

int sms_set_str(char *in, char *out)
{
	char tmp[512];
	int i,j,k;
	int outlen;
	int outhexlen = 0;

	int inlen = strlen(in);

	for (i = 0, k = 0; i < inlen;) {
		outlen = sms_decode_gb18030(in + i, 2, tmp);
		i += 2;

		outhexlen += outlen;
		for (j = 0; j < outlen; j ++) {
			out[k++] = tmp[j];
		}
	}

	out[outhexlen] = '\0';

	return outhexlen;
}

static int sms_encode_gb18030(char *in, size_t inlen, char *out, size_t outlen)
{
	char tmp[512];
	int outleft;
	int i;
	int j;


	tmp[0] = 0xFF;
	tmp[1] = 0xFE;
	j = 2 + inlen;

	for (i = 0; i < inlen; i += 2) {
		tmp[i + 2] = in[i + 1];
		tmp[i + 3] = in[i];
	}


	outleft = unicode_2_gb18030(tmp, j, out, outlen);
	if (outleft <= 0) {
		return -1;
	}

	out[outlen-outleft] = '\0';

	return outlen - outleft;
}

int sms_get_str(char *in, size_t inlen, char *out, size_t outlen)
{
	return sms_encode_gb18030(in, inlen, out, outlen);
}

int gsm_text2sm_event(struct gsm_modul *gsm, char *sms_info) 
{
	char tmp[256];

	char *p;
	char *ret;
	int language;
	char *sms_body;

	if (NULL == sms_info) {
		return -1;
	}

	p = sms_info;

	/* get %CMT */
	ret = gsm_get_dot_string(p, tmp, ':'); // +CMT:
	if (!ret) {
		return -1;
	}

	/* get callerID */
	ret = gsm_get_dot_string(ret, gsm->ev.sm_received.sender, ','); // callerID
	if (!ret) {
		return -1;
	}

	/* get year */
	ret = gsm_get_dot_string(ret, tmp, ','); // year
	if (!ret) {
		return -1;
	}

	/* get month */
	ret = gsm_get_dot_string(ret, tmp, ','); // month
	if (!ret) {
		return -1;
	}

	/* get day */
	ret = gsm_get_dot_string(ret, tmp, ','); // day
	if (!ret) {
		return -1;
	}

	/* get hour */
	ret = gsm_get_dot_string(ret, tmp, ','); // hour
	if (!ret) {
		return -1;
	}

	/* get minute */
	ret = gsm_get_dot_string(ret, tmp, ','); // minute
	if (!ret) {
		return -1;
	}

	/* get language format */
	ret = gsm_get_dot_string(ret, tmp, ','); // format
	if (!ret) {
		return -1;
	}
	language = atoi(tmp);

	/* get length */
	ret = gsm_get_dot_string(ret, tmp, ','); // length
	if (!ret) {
		return -1;
	}
	gsm->ev.sm_received.len = atoi(tmp);

	/* get length */
	ret = gsm_get_dot_string(ret, tmp, ','); // prt
	if (!ret) {
		return -1;
	}

	/* get length */
	ret = gsm_get_dot_string(ret, tmp, ','); // prv
	if (!ret) {
		return -1;
	}

	/* get length */
	ret = gsm_get_dot_string(ret, tmp, '\r'); // type
	if (!ret) {
		return -1;
	}

	sms_body = ret + 1;

	/* get SMS body */
	if (6 == language) {
		sms_get_str(sms_body, gsm->ev.sm_received.len, gsm->ev.sm_received.text, 255);
	} else {
		strncpy(gsm->ev.sm_received.text, sms_body, gsm->ev.sm_received.len);
		gsm->ev.sm_received.text[gsm->ev.sm_received.len] = '\0';
	}

	return 0;
}


