#ifndef _OPVXG4XX_H_
#define _OPVXG4XX_H_

/* debug level */
#define DEBUG_G4_NONE     0x0000
#define DEBUG_G4_SYS      0x0001
#define DEBUG_G4_MSG      0x0002
#define DEBUG_G4_STATE    0x0004
#define DEBUG_G4_ZAP      0x0008
#define DEBUG_G4_SER      0x0010
#define DEBUG_G4_CODEC    0x0020
#define DEBUG_G4_I2C      0x0040

/* external device address */
#define UART0_BASE      0x000004c0
#define UART1_BASE      0x000004e0
#define UART2_BASE      0x00000500
#define UART3_BASE      0x00000520
#define PIO0_BASE       0x000005e0
#define SWAP_MEM_BASE   0x00000000
#define IRQ_BASE        0x0000c000
#define I2C0_BASE       0x00000540
#define I2C1_BASE       0x00000560
#define I2C2_BASE       0x00000580
#define I2C3_BASE       0x000005a0
#define TDM0_BASE       0x00000400
#define BUS_FREQ        66000000
#define DP_MEM_BASE     0x00001000
#define I2_PRESCALE     ((BUS_FREQ/5/100000)) /* prescale = busfreq/(5*I2cFreq), now we use 100k I2c Freq; */

/* I2C control and flag bits */
#define I2C_FREQ_LOW        0x0
#define I2C_FREQ_HIGH       0x1
#define I2C_CTRL            0x2
#define I2C_TX              0x3
#define I2C_RX              0x3
#define I2C_CMD             0x4
#define I2C_STATUS          0x4

#define I2C_CTRL_ENABLE     0X80    // enable I2C core
#define I2C_CTRL_IRQEN      0X40    // enable interrupt
#define I2C_CMD_STA         0x80    // generate(repeated) start condition
#define I2C_CMD_STO         0x40    // generate stop condition
#define I2C_CMD_RD          0x20    // read from slave
#define I2C_CMD_WR          0x10    // write to slave
#define I2C_CMD_ACK         0x08    // ACK, when a receiver, sent ACK(0) or NACK(1)
#define I2C_CMD_IACK        0x01    // Interrupt acknowledge, set 1 clear.
#define I2C_STATUS_RXACK    0x80    // Received acknowledge from slave. 0= Acknowledge received, 1= no 
#define I2C_STATUS_BUSY     0x40    // I2C bus busy
#define I2C_STATUS_AL       0x20    // Arbitration lost     
#define I2C_STATUS_TIP      0x02    // Trsnasfer in progress
#define I2C_STATUS_IF       0x01    // interrupt is pending
#define I2C_TIMEOUT         0x64    /* wait 10 ms before i2c time out; */

#define PIO0_DTR        0
#define PIO0_DCD        4
#define PIO0_RI         8
#define PIO0_STATUS     12
#define PIO0_RST        16
#define PIO0_PWR        20
#define PIO0_PORTSEL    24
#define PIO0_SLOTUSE    28

/* PIO0 control and flag bits */
#define PIO0_DTR_MASK       (0xfu<<PIO0_DTR)
#define PIO0_DCD_MASK       (0xfu<<PIO0_DCD)
#define PIO0_RI_MASK        (0xfu<<PIO0_RI)
#define PIO0_STATUS_MASK    (0xfu<<PIO0_STATUS)
#define PIO0_RST_MASK       (0xfu<<PIO0_RST)
#define PIO0_PWR_MASK       (0xfu<<PIO0_PWR)
#define PIO0_PORTSEL_MASK   (0xfu<<PIO0_PORTSEL)
#define PIO0_SLOTUSE_MASK   (0xfu<<PIO0_SLOTUSE)

#ifndef PCI_VENDOR_ID_OPENVOX
#define PCI_VENDOR_ID_OPENVOX       0x1B74
#define PCI_DEVICE_ID_OPENVOX       0x0100
#define PCI_SUBSYSTEM_ID_OPENVOX    0x0104
#endif

typedef struct g4_gpio
{
    unsigned int base;      /* base address */
    unsigned int dir;       /* io direction register value */
    unsigned int io;        /* previous io register value */
} g4_gpio;

#define G4_FIFO_SIZE        16
#define G4_SPANS            4
#define G4_SER_BUF_SIZE     1024
#define G4_CHAN_NUM 2

typedef struct g4_span {
    /* zaptel resources */
    struct zt_span span;
    struct zt_chan chans[G4_CHAN_NUM];
    struct g4xx_card *card;
    /* more zaptel stuff */
    unsigned int usecount;
    int spantype;
    int spanflags;
} g4xx_span;

typedef struct g4_card {
    spinlock_t lock;
    unsigned char power[G4_SPANS];
    unsigned int dtr_on_off;
    unsigned int fwversion;
    char *card_name;
    int rst_flag;
    int cardID;
    int dead;
    unsigned char cardno;
    unsigned int irq;
    unsigned int iomem;
    void *pci_io;
    unsigned long pci_io_phys;
    unsigned long iomem_size;
    struct g4_span spans[G4_SPANS];
    unsigned int g4spans;
    struct pci_dev *pcidev;
    unsigned int type;
    unsigned char b_rx_buf[G4_SPANS][ZT_CHUNKSIZE];
    unsigned char b_tx_buf[G4_SPANS][ZT_CHUNKSIZE];
    unsigned char d_rx_buf[G4_SPANS][G4_SER_BUF_SIZE];
    unsigned char d_tx_buf[G4_SPANS][G4_FIFO_SIZE];

    unsigned char ser_rx_buf[G4_SPANS][G4_SER_BUF_SIZE];
    unsigned char ser_tx_buf[G4_SPANS][G4_SER_BUF_SIZE];
    unsigned short ser_rx_idx[G4_SPANS];
    unsigned short ser_tx_idx[G4_SPANS];

    int d_rx_recv[G4_SPANS];
    
    g4_gpio gpio0;

} g4_card;


#endif /*_OPVXG4XX_H_*/
